#include "filtering.h"

#define MAX_MASKED_REGIONS 128

#define DO_MASK

/**
 * Resets a filter, by erasing the masked zones (nothing is aligned), resetting the crt_hit field,
 * and setting a new index if the index parameter is not null
 * @param filter The filter to reset
 * @param index The new index. If this parameter is NULL, the previous index is kept.
 */
inline void reg_filter__reset(RegionFilterType* filter) {
  // wipe the map clean
  if (filter->masked_counter >= MAX_MASKED_REGIONS) {
    memset(filter->mask, REGION_MASK_AVAILABLE, filter->size * sizeof(REGION_FILTER_MASK_TYPE));
    memset(filter->mask_positions, 0x0, MAX_MASKED_REGIONS * sizeof(int));
  }
  else {
    int i;
    for (i = 0; i < filter->masked_counter; ++i) {
      int to_unmask = min(filter->window_size, filter->size - filter->mask_positions[i]);
      memset(filter->mask + filter->mask_positions[i], REGION_MASK_AVAILABLE, to_unmask * sizeof(REGION_FILTER_MASK_TYPE));
      filter->mask_positions[i] = 0;
    }
  }
  filter->masked_counter = 0;
}

/**
 * Initialize a filter, for a given reference database, that is based on a given index.
 * @param filter
 * @param db
 * @param index
 */
inline void reg_filter__init(RegionFilterType* filter, ReferenceDBType* db, const int window_size) {
  filter->size = db->size;
  filter->window_size = window_size;
  SAFE_FAILURE__ALLOC(filter->mask, filter->size, REGION_FILTER_MASK_TYPE);
  SAFE_FAILURE__ALLOC(filter->mask_positions, MAX_MASKED_REGIONS, int);
  filter->masked_counter = MAX_MASKED_REGIONS;
  reg_filter__reset(filter);
}

/**
 * Create a filter, for a given reference database, that is based on a given index.
 * @param db
 * @param index
 * @return the address of the created filter
 */
RegionFilterType* reg_filter__create(ReferenceDBType* db, const int window_size) {
  RegionFilterType* filter;
  SAFE_FAILURE__ALLOC(filter, 1, RegionFilterType);
  reg_filter__init(filter, db, window_size);
  return filter;
}

/**
 * Finds the next hit of a key in the index. If the next hit is in an already aligned zone,
 * the search continues until the hit is in an available zone or there are no more hits
 * for this key in the index.
 * @param filter
 * @param key
 * @return The next hit that is not in an already aligned zone, or -1 if there is no such hit.
 */
inline int reg_filter__next_hit(RegionFilterType* filter, IndexType* index, const int key, const int _ogtn){

#ifdef DO_MASK
  if (filter->masked_counter < MAX_MASKED_REGIONS) {
    int i = 0;
    int a;
  retry_filter:
    a = index__get_next_hit(index, key, _ogtn);
    if (a < 0)
      return a;
    while (i < filter->masked_counter) {
      if  (a < filter->mask_positions[i])
        return a;
      else
        if (a < filter->mask_positions[i]+filter->window_size)
          goto retry_filter;
        else
          i++;
    }
    return a;
  } else {
    int a;
    do {
      a = index__get_next_hit(index, key, _ogtn);
    } while (a >= 0 && filter->mask[a] == REGION_MASK_ALIGNED);
    /* crt_hit == -1  =>  no more hits for this key, stop */
    /* mask[filter->crt_hit] == AVAILABLE  =>  this zone has not already been aligned, report this hit */
    return a;
  }
#else
  return index__get_next_hit(index, key, _ogtn);
#endif
}
/**
 * Finds the next hit of a key in the index. If the next hit is in an already aligned zone,
 * the search continues until the hit is in an available zone or there are no more hits
 * for this key in the index.
 * @param filter
 * @param key
 * @return The next hit that is not in an already aligned zone, or -1 if there is no such hit.
 */
inline int reg_filter__extern_next_hit(RegionFilterType* filter, IndexType* index, const int key, int* hit_pointer){
#ifdef DO_MASK
  if (filter->masked_counter < MAX_MASKED_REGIONS) {
    int i = 0;
    int a;
  retry_filter:
    a = index__get_extern_next_hit(index, key, hit_pointer);
    if (a < 0)
      return a;
    while (i < filter->masked_counter) {
      if  (a < filter->mask_positions[i])
        return a;
      else
        if (a < filter->mask_positions[i]+filter->window_size)
          goto retry_filter;
        else
          i++;
    }
    return a;
  } else {
    int a;
    do {
      a = index__get_extern_next_hit(index, key, hit_pointer);
    } while (a >= 0 && filter->mask[a] == REGION_MASK_ALIGNED);
    /* crt_hit == -1  =>  no more hits for this key, stop */
    /* mask[filter->crt_hit] == AVAILABLE  =>  this zone has not already been aligned, report this hit */
    return a;
  }
#else
  return index__get_extern_next_hit(index, key, hit_pointer);
#endif
}

/**
 * Finds the next <tt>count</tt> hits of a key in the index, avoiding already matched zones.
 * This is useful for allowing the quick alignment filter perform several alignments simultaneously.
 * @param filter
 * @param key
 * @param dest a vector where the hit positions are placed
 * @param count the number of hits to find
 * @return The last hit placed in dest, or -1 if there is no such hit.
 */
inline int reg_filter__next_hits(RegionFilterType* filter, IndexType* index, const int key, int* dest, const int count, const int _ogtn) {
  int i, last = -1, return_ = -1;
  for (i = 0; i < count; ++i) {
    // get next hit
    if ((return_ = dest[i] = reg_filter__next_hit(filter, index, key, _ogtn)) < 0) {
      // if there are no more hits, break and fill with the last found hit
      break;
    } else {
      // memorize the last hit
      last = dest[i];
    }
  }
  for (; i < count; ++i) {
    dest[i] = last;
  }
  return return_;
}
/**
 * Finds the next <tt>count</tt> hits of a key in the index, avoiding already matched zones.
 * This is useful for allowing the quick alignment filter perform several alignments simultaneously.
 * @param filter
 * @param key
 * @param dest a vector where the hit positions are placed
 * @param count the humber of hits to find
 * @return The last hit placed in dest, or -1 if there is no such hit.
 */
inline int reg_filter__extern_next_hits(RegionFilterType* filter, IndexType* index, const int key, int* dest, const int count, int* hit_pointer){
  int i, last = -1, return_ = -1;
  for (i = 0; i < count; ++i) {
    // get next hit
    if ((return_ = dest[i] = reg_filter__extern_next_hit(filter, index, key, hit_pointer)) < 0) {
      // if there are no more hits, break and fill with the last found hit
      break;
    } else {
      // memorize the last hit
      last = dest[i];
    }
  }
  for (; i < count; ++i) {
    dest[i] = last;
  }
  return return_;
}

/**
 * Masks the part of the reference that was successfully aligned, so that the next
 * hit search can avoid it.
 * @param filter
 * @param alignment The alignment of the reference part
 * @param reference_start The position of the aligned part in the reference
 * @return 0 if the masking was successful, -1 if the alignment data is wrong
 * (alignment not performed yet in the given structure).
 */
inline int reg_filter__mask(RegionFilterType* filter, AlignmentType* alignment, const int reference_start) {
#ifdef DO_MASK
  if (alignment->best_j0 < 0) {
    if (alignment->best_j < 0) {
      // not yet aligned, can't mask
      return -1;
    }
    alignment__traceback(alignment);
  }
  //mask
  int to_unmask = min(filter->window_size, filter->size - reference_start - alignment->best_j0);
  memset(filter->mask + reference_start + alignment->best_j0, REGION_MASK_ALIGNED, to_unmask * sizeof(REGION_FILTER_MASK_TYPE));
  //mark the mask
  if (filter->masked_counter < MAX_MASKED_REGIONS) {
    int pm = reference_start + alignment->best_j0;
    filter->mask_positions[filter->masked_counter] = pm;
    /* keep sort order (quadratic but local) */
    int i = filter->masked_counter;
    while ( i > 0 && filter->mask_positions[i-1] > pm) {
      filter->mask_positions[i] = filter->mask_positions[i-1];
      i--;
    }
    filter->mask_positions[i] = pm;
  }
  ++ filter->masked_counter;
#endif
  return 0;
}

/**
 * Cleans up the memory allocated by a filter
 * @param filter
 */
inline void reg_filter__destroy(RegionFilterType* filter) {
  free(filter->mask_positions);
  free(filter->mask);
}
