#ifndef _INDEX_H_
#define _INDEX_H_

#ifdef _OPENMP
#include <omp.h>
#endif

/* #define DEBUG_INDEX */

#include <math.h>

#include "seed.h"
#include "load_data.h"
#include "reads_data.h"
#include "reference_data.h"


#define DEFAULT_ACCEPTED_STDEV_DISTANCE 100

#ifdef _OPENMP
extern int MAX_THREADS;

#define INDEX_CRT_HIT__DECLARE  int *crt_hit;
#define INDEX_CRT_HIT__INIT     int t; \
      SAFE_FAILURE__ALLOC(index->crt_hit, MAX_THREADS, int);\
      for (t = 0; t < MAX_THREADS; ++t) {index->crt_hit[t] = -1;}
#define INDEX_CRT_HIT__RESET(_t) {index->crt_hit[_t] = -1;}
#define INDEX_CRT_HIT__REF      index->crt_hit[_ogtn]
#define INDEX_CRT_HIT__DESTROY  free(index->crt_hit);
#define INDEX_HIT_COUNTS_INIT   SAFE_FAILURE__ALLOC(index->hits_counts, MAX_THREADS, int);
#define INDEX_HIT_COUNTS_INC(_t) index->hits_counts[_t]++;
#define INDEX_HIT_COUNTS_DESTROY free(index->hits_counts);
#else
#define INDEX_CRT_HIT__DECLARE int crt_hit
#define INDEX_CRT_HIT__INIT    index->crt_hit = -1;
#define INDEX_CRT_HIT__RESET(_t) index->crt_hit = -1;
#define INDEX_CRT_HIT__REF     index->crt_hit
#define INDEX_CRT_HIT__DESTROY ;
#define INDEX_HIT_COUNTS_INIT  ;
#define INDEX_HIT_COUNTS_INC(_t) index->hits_count++;
#define INDEX_HIT_COUNTS_DESTROY ;
#endif

/**
 * @struct IndexType
 *
 */


typedef struct IndexType {
  /** reference to the genome sequence db */
  ReferenceDBType *db;
  /** reference to the seed used to build the index */
  SeedType* seed;
  /** "seedcode" indexed array : gives the first "position" of
   *  a given "seedcode"
   */
  int* first;
  /** "position" indexed array : gives the next "position" of
   *   a "seedcode" that also occurs at this "position"
   *   (linked list principle)
   *   @see first
   */
  int* map;
  INDEX_CRT_HIT__DECLARE;
  int index_size;
  int map_size;
#ifdef DEBUG_INDEX
#ifdef _OPENMP
  int* hits_counts;
#endif
  int hits_count;
#endif
} IndexType;

/**
 * Builds the index of a genome according to a given seed pattern
 * @param db
 * @param seed
 * @return The address of the created index
 */
IndexType* index__build_reference(ReferenceDBType* db, SeedType* seed);

/**
 * Reset the index "pointer" to the current hit
 */
inline void index__reset_crt_hit(IndexType *index, const int _ogtn);
/**
 * Reset the index external "pointer" to the current hit
 */
inline void index__reset_extern_crt_hit(int *hit_pointer);
/**
 * A hit in a reference index
 * @param index
 * @param key
 * @return The position -- in the reference database -- of the subsequence
 * corresponding to the key
 */
int index__get_next_hit(IndexType *index, const int key, const int _ogtn);
/**
 * A hit in a reference index (as field external to the index structure)
 * @param index
 * @param key
 * @return The position -- in the reference database -- of the subsequence
 * corresponding to the key
 */
int index__get_extern_next_hit(const IndexType *index, const int key, int* hit_pointer);


inline int index__get_crt_hit (const IndexType *index, const int key, const int _ogtn);

inline int index__get_extern_crt_hit (const IndexType *index, const int key, const int *hit_pointer);


/**
 * A hit in a read index
 * @param index
 * @param key
 * @param read_id Address where the number of the read where the next
 * occurrence was found is placed
 * @param read_pos Address where the position within the read where the next
 * occurrence was found is placed
 * @return The position -- wrt the index -- of the subsequence
 * corresponding to the key
 */
int index__get_hit_read(IndexType *index, const int key, int* read_id, int* read_pos, const int _ogtn);

/**
 * How many hits were found in an index during a certain run
 *
 * @param index
 * @return number of hits
 */
inline int index__get_hits_count(IndexType* index);

/**
 * Destroy an indexed
 * @param index
 */
void index__destroy(IndexType* index);

void index__debug_print(const IndexType* index, const char* filename);

#endif /* _INDEX_H_ */
